"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

This file holds code and data class definitions for things peculiar
to the MISSIONS directory. These "things" aren't quite entities, as
they lack associated string resource fields and/or nicknames.
Instead they "belong to" a composite Entity, like a Base or Faction.
"""
from typing import Tuple, List, Dict, Optional

from dataclassy import dataclass

from .formats import ini
from . import cached, paths


@cached
def get_faction_props() -> Dict[str, 'FactionProps']:
    """Produce a dictionary of faction nicknames to their associated FactionProps object."""
    props = ini.sections(paths.construct_path('DATA/MISSIONS/faction_prop.ini'))
    return {p['affiliation']: FactionProps(**p) for p in props['factionprops']}


@cached
def get_mbases() -> Dict[str, 'MBase']:
    """Produce a dictionary of base nicknames to their associated MBase entry. Each MBase section's associated
    MVendor, BaseFaction and GF_NPC sections are folded into attributes of the resulting MBase object.

    Implementation note: because the ordering of these sections is guaranteed (breaking this rule will crash
    Freelancer), it is safe to assume that base is defined in subsequent code paths."""
    sections = ini.group(paths.construct_path('DATA/MISSIONS/mbases.ini'), fold_sections=False)

    bases = []

    for name, contents in sections:
        if name == 'mbase':
            base = MBase(**contents[0])
            bases.append(base)
        elif name == 'mvendor':
            # noinspection PyUnboundLocalVariable
            base.vendors.append(MVendor(**contents[0]))
        elif name == 'basefaction':
            base.factions.extend(BaseFaction(**faction) for faction in contents)
        elif name == 'gf_npc':
            base.npcs.extend(GF_NPC(**npc) for npc in contents if 'nickname' in npc)
        elif name == 'mroom':
            base.rooms.extend(MRoom(**room) for room in contents)

    return {b.nickname: b for b in bases}


@dataclass
class MBase:
    """A "mission base" section, found in mbases.ini. Provides additional attributes to the base that nickname refers
    to. The optional fields of this class, as with the following classes, are based on my observations of Discovery
    Freelancer. There are likely more fields which Freelancer does not strictly require."""
    nickname: str
    local_faction: str
    diff: int
    msg_id_prefix: Optional[str] = None

    vendors: List['MVendor'] = []
    factions: List['BaseFaction'] = []
    npcs: List['GF_NPC'] = []
    rooms: List['MRoom'] = []


@dataclass
class MVendor:
    """Found in mbases.ini, this section describes the preceding base's mission vendor, aka the "jobs board"."""
    num_offers: Tuple[int, int]


@dataclass
class BaseFaction:
    """Found in mbases.ini, this section describes an individual NPC present on the preceding base."""
    faction: str
    weight: int
    offers_missions: bool = False
    mission_type: Optional[Tuple[str, float, float, int]] = None
    npc: List[str] = []


# noinspection PyPep8Naming
@dataclass
class GF_NPC:
    """Found in mbases.ini, this section describes an individual NPC present on the preceding base."""
    nickname: str
    body: Optional[str] = None
    head: Optional[str] = None
    lefthand: Optional[str] = None
    righthand: Optional[str] = None
    individual_name: int
    affiliation: str
    voice: str
    misn: Tuple[str, float, float] = []
    room: Optional[str] = None
    bribe: List[Tuple[str, int, int]] = []
    rumor: List[Tuple[str, str, int, int]] = []
    know: Optional[Tuple[int, int, int, int]] = None
    knowdb: Optional[str] = None
    rumorknowdb: Optional[str] = None
    accessory: Optional[str] = None
    base_appr: Optional[str] = None
    rumor_type2: Optional[Tuple[str, str, int, int]] = None


@dataclass
class MRoom:
    """Found in mbases.ini, this section describes a particular room in the preceding base."""
    nickname: str
    character_density: int = 0
    fixture: List[Tuple[str, str, str, str]] = []


@dataclass
class FactionProps:
    """The FactionProps section, found in faction_prop.ini, defines much of the behaviour of a faction's NPCs in
    space."""
    affiliation: str
    legality: str
    nickname_plurality: str
    msg_id_prefix: str
    jump_preference: str
    npc_ship: List[str] = []
    voice: List[str] = []
    mc_costume: str
    space_costume: List[Tuple[str, str, str]] = []
    firstname_male: Optional[Tuple[int, int]] = None
    firstname_female: Optional[Tuple[int, int]] = None
    lastname: Tuple[int, int]
    rank_desig: Tuple[int, int, int, int, int]
    formation_desig: Tuple[int, int]
    large_ship_desig: Optional[int] = None
    large_ship_names: Tuple[int, int] = None
    scan_for_cargo: List[Tuple[str, int]] = []
    scan_announce: bool = False
    scan_chance: float = 0.0
    formation: List[Tuple[str, str]]



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

This file provides an interactive prompt for working with flint
(and therefore a Freelancer installation). Run with `python -i`,
e.g. `python -i interactive.py "<freelancer_dir>"`.

If running in PyCharm, ensure "Emulate terminal in output console"
is enabled in the run/debug configuration.
"""
import argparse
import os
import subprocess
import sys
import warnings

# if not running in interactive mode, make it so
# this trick allows the very handy "python -m flint.interactive" to bring up the interactive shell mode
if not sys.flags.interactive:
    subprocess.call([sys.executable, '-i', *sys.argv])
    exit(0)

# make flint available for import
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), '../..'))

# override interpreter prompts
sys.ps1 = '(flint as fl) >>> '
sys.ps2 = '              ... '

# parse command line arguments
parser = argparse.ArgumentParser(description='flint, a parser for Freelancer and its formats')
parser.add_argument('freelancer_dir', help='Path to a working Freelancer install directory')
parser.add_argument('-d', '--discovery', help='Whether Freelancer dir is Discovery modded.', action='store_true')
parser.add_argument('-s', '--silent', help='Silence all warnings', action='store_true')
try:
    arguments = parser.parse_args()
except SystemExit:
    os._exit(1)

# configure warnings
warnings.simplefilter('ignore' if arguments.silent else 'always')

import flint as fl
fl.paths.set_install_path(arguments.freelancer_dir, arguments.discovery)



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
from typing import Dict, List, Tuple, Optional, cast

from dataclassy import Internal

from . import Entity, EntitySet
from .. import paths
from ..formats import utf


class Good(Entity):
    """A Good is the physical (transferable) representation of something that can be bought or sold. A good maps the
    abstract definition of commodities, equipment and ships to something that is tradeable."""
    ids_info = None
    ids_name = None
    item_icon: Optional[str] = None  # path to icon, relative to DATA
    price: int  # the default price for this good, pre market multiplier

    def icon_path(self) -> str:
        """The absolute path to the .3db file containing this item's icon."""
        return paths.construct_path('DATA', self.item_icon or self.DEFAULT_ICON)

    def icon(self) -> bytes:
        """This good's icon in TGA format."""
        return utf.extract(self.icon_path(), 'MIP0')

    def market(self) -> Dict[bool, Dict['Base', int]]:
        """The market for this Good, i.e. the Bases it is bought and sold on and their prices."""
        return routines.get_markets()[self]

    def sold_at(self) -> Dict['Base', int]:
        """A dict of bases that sell this good of the form {base_nickname: price}."""
        return self.market()[True]

    def bought_at(self) -> Dict['Base', int]:
        """A dict of bases that buy this good of the form {base_nickname: price}."""
        return {**self.market()[False], **self.sold_at()}

    def price_at(self, base: 'Base') -> int:
        return self.bought_at()[base]

    DEFAULT_ICON = 'EQUIPMENT/MODELS/COMMODITIES/NN_ICONS/blank.3db'


class EquipmentGood(Good):
    """The good of a piece of equipment."""
    equipment: str  # nickname of the good this equipment represents.
    combinable: bool

    def equipment_(self) -> 'Equipment':
        """The Equipment entity this good refers to."""
        return routines.get_equipment().get(self.equipment)


class CommodityGood(EquipmentGood):
    """The good of a commodity. This is the place where the distinction between equipment and commodities is made."""
    good_sell_price: float
    bad_buy_price: float
    bad_sell_price: float
    good_buy_price: float
    shop_archetype: str

    def commodity(self) -> 'Commodity':
        """The Commodity entity this good refers to."""
        return cast(Commodity, self.equipment_())


class ShipHull(Good):
    """The hull of a ship, meaning a ship with no equipment mounted."""
    ship: str  # nickname of a Ship

    def ship_(self) -> 'Ship':
        """The Ship that uses this hull."""
        return routines.get_ships().get(self.ship)


class ShipPackage(Good):
    """A ship "package" is the form a ship is buyable in. As a kind of "composite" good, it does not have an inherent
    `price` attribute, unlike other goods. Its cost is the sum of the hull and all "addons", which are the equipment
    mounted on the ship when purchased."""
    price = 0
    hull: str  # nickname of a ShipHull
    addon: Internal[List[Tuple[str, str, int]]]  # tuple of the form (equipment nickname, hardpoint nickname, ?)

    def hull_(self) -> ShipHull:
        """The ShipHull entity of this package's hull."""
        return routines.get_goods()[self.hull]

    def ship(self) -> 'Ship':
        """The Ship this package represents."""
        return self.hull_().ship_()

    def cost(self) -> int:
        """The cost of this ship package when bought. Note that this is distinct from Good's _price_, which ShipPackage
        lacks. Instead a ship's cost at the vendor is the sum of the prices of its hull and default equipment
        ("addons")."""
        return self.hull_().price  # todo: include cost of addons

    def equipment(self) -> EntitySet['Equipment']:
        """The set of equipment included in this package."""
        equipment = routines.get_equipment()
        return EntitySet(equipment[n] for n,  *_ in self.addon if n in equipment)


from .. import routines
from .equipment import Equipment, Commodity


"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

Interface-related functions, such as routines for translating RDL.
"""
import xml.etree.ElementTree as xml


def rdl_to_html(rdl: str) -> str:
    """Translate RDL to HTML. Currently this uses a crude lookup table. In future I want to replace this with
    proper interpretation of the XML."""
    result = rdl
    for rdl_tag, html_tag in RDL_TO_HTML.items():
        result = result.replace(rdl_tag, html_tag)
    return result


def rdl_to_plaintext(rdl: str) -> str:
    """Translate RDL to plaintext, stripping all tags and replacing <PARA/> with a newline."""
    rdl = rdl.replace('<PARA/>', '\n').replace('</PARA>', '')
    tree = xml.fromstring(rdl)
    return xml.tostring(tree, encoding='unicode', method='text')


def html_to_rdl(html: str) -> str:
    """Translate HTML to RDL. See the docstring for `rdl_to_html` for more information."""
    result = html
    for rdl_tag, html_tag in RDL_TO_HTML.items():
        result = result.replace(html_tag, rdl_tag)
    return result


# A lookup table mapping RDL (Render Display List) tags to HTML(4). Freelancer, to my eternal horror, uses these for
# formatting for strings inside these resource DLLs. Based on work by adoxa and cshake.
# More information can be found in this thread: <https://the-starport.net/modules/newbb/viewtopic.php?&topic_id=562>
RDL_TO_HTML = {
    '<TRA data="1" mask="1" def="-2"/>':           '<b>',  # bold
    '<TRA bold="true"/>':                          '<b>',  # rare bold
    '<TRA data="0" mask="1" def="-1"/>':           '</b>',  # un-bold
    '<TRA data="2" mask="3" def="-3"/>':           '<i>',  # italic 1
    '<TRA data="0" mask="3" def="-1"/>':           '</i>',  # un-italic 1
    '<TRA data="98" mask="-29" def="-3"/>':        '<i>',  # italic 2
    '<TRA data="96" mask="-29" def="-1"/>':        '</i>',  # un-italic 2
    '<TRA data="2" mask="2" def="-3"/>':           '<i>',  # italic 3
    '<TRA data="0" mask="2" def="-1"/>':           '</i>',  # un-italic 3
    '<TRA data="5" mask="5" def="-6"/>':           '<b><u>',  # (bold, underline) 1
    '<TRA data="0" mask="5" def="-1"/>':           '</b></u>',  # un-(bold, underline) 1
    '<TRA data="5" mask="7" def="-6"/>':           '<b><u>',  # (bold, underline) 2
    '<TRA data="0" mask="7" def="-1"/>':           '</b></u>',  # un-(bold, underline) 2
    '<TRA data="65280" mask="-32" def="31"/>':     '<font color="red">',  # red
    '<TRA data="96" mask="-32" def="-1"/>':        '</font>',  # un-colour
    '<TRA data="65281" mask="-31" def="30"/>':     '<b><font color="red">',  # (bold, red)
    '<TRA data="96" mask="-31" def="-1"/>':        '</b></font>',  # un-(bold, red)
    '<TRA data="-16777216" mask="-32" def="31"/>': '<font color="blue">',  # blue
    '<PARA/>':                                     '<p>',  # newline
    '</PARA>':                                     '</p>',
    '<JUST loc="left"/>':                          '<p align="left">',  # newline with left-aligned text
    '<JUST loc="center"/>':                        '<p align="center">',  # newline with centred text
    '\xa0':                                        '&nbsp;',  # non-breaking space, is often present after the title
    '<RDL>':                                       '',  # seemingly meaningless tags...
    '</RDL>':                                      '',
    '<TEXT>':                                      '',
    '</TEXT>':                                     '',
    '<PUSH/>':                                     '',
    '<POP/>':                                      '',
    '<?xml version="1.0" encoding="UTF-16"?>':     '',  # xml header; removed for neatness
}



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

This file provides a simple, speedy parser for Freelancer-style
INI files, which are used to store all information about the game
world.

Freelancer actually stores INIs in a compressed binary-INI (BINI)
format, but will accept text INIs happily. This is therefore the
format most used by mods as it facilitates editing.

This file is intended to be the main interface for accessing INI
*and* BINI functions, as it contains higher-level functions as well
as logic for checking whether a .ini file is an INI or a BINI.
"""
from typing import Union, List, Dict, Any, Tuple
from collections import defaultdict
from functools import lru_cache
import concurrent.futures
import itertools
import warnings

from . import bini


@lru_cache(64)
def sections(paths: Union[str, Tuple[str]], fold_sections=False, fold_values=True) -> Dict[str, Any]:
    """Parse the Freelancer-style INI file(s) at `paths` and group sections of the same name together.

    THe result is a dict mapping a section name to a list of dictionaries representing the contents of each section
    "instance". If `fold_sections` is true, this list will be "folded" into one dict for sections with only one
    instance.

    If `fold_values` is true (the default), the same logic applies to entries and their values: if an entry is only
    defined once in a section, its value is "folded" into a primitive (a float, int, bool or string) rather than being a
    list."""
    return fold_dict(parse(paths, fold_values), fold_sections)


def parse(paths: Union[str, Tuple[str]], fold_values=True) -> List[Tuple[str, Dict[str, Any]]]:
    """Parse an INI file, or a collection of INIs, to a list of tuples of the form (section_name, section_contents),
    where section_contents is a dict of the entries in that section. If fold_values is true (the default), the
    entries dict will be "folded" (see the docstring for `fold_dict`)."""
    if isinstance(paths, str):  # accept both single paths and tuples of paths
        paths = [paths]

    with concurrent.futures.ThreadPoolExecutor() as executor:
        sections_ = itertools.chain(*executor.map(parse_file, paths))

    return [(name, fold_dict(entries, fold_values)) for name, entries in filter(None, sections_)]


def group(paths: Union[str, Tuple[str]], fold_sections=True, fold_values=True):
    """Similar to `parse` but groups contiguous sequences of the same section name together."""
    groups = itertools.groupby(parse(paths, fold_values), key=lambda pair: pair[0])  # group by section name
    return [next(iter(fold_dict(contents, fold_sections).items())) for key, contents in groups]


def parse_file(path: str):
    """Takes a path to an INI or BINI file and outputs a list of tuples containing a section name and a list of tuples
    of entry/value pairs."""
    if bini.is_bini(path):
        return bini.parse_file(path)
    with open(path, encoding='windows-1252') as f:
        contents = f.read().lower()  # files are case insensitive
    contents.replace(DELIMITER_COMMENT + SECTION_NAME_START, '')  # delete commented section markers
    return list(map(parse_section, contents.split(SECTION_NAME_START)))


def parse_section(section: str):
    """Takes a raw section string (minus the [) and outputs a tuple containing the section name and a list of tuples
    of entry/value pairs. If the section is invalid, an empty tuple will be returned."""
    section_name, delimiter, entries = section.partition(SECTION_NAME_END)
    if not delimiter or (DELIMITER_COMMENT in section_name):
        return ()
    try:
        return section_name, list(map(parse_entry, entries.splitlines()))
    except ValueError as e:  # an entry with a syntax error invalidates the whole section
        warnings.warn(f"Couldn't parse line in section {section_name!r}; {e.args[0]}")
        return ()


def parse_entry(entry: str):
    """Takes an entry string consisting of a delimiter separated key/value pair and outputs a tuple of the
    name and value. If the entry is invalid, an empty tuple will be returned."""
    entry = entry.split(DELIMITER_COMMENT, 1)[0].replace(' ', '').replace('\t', '')  # remove comments and whitespace
    key, delimiter, value = entry.partition(DELIMITER_KEY_VALUE)
    if not delimiter:  # if this isn't a valid entry line after all
        return ()
    return key, parse_value(value)


def parse_value(entry_value: str) -> Union[Any, Tuple]:
    """Parse an entry value (consisting either of a string, int or float or a tuple of such) using and return it as a
    Python object."""
    return tuple(map(auto_cast, entry_value.split(','))) if ',' in entry_value else auto_cast(entry_value)


def auto_cast(value: str) -> Any:
    """Interpret and coerce a string value to a Python type. If the value cannot be interpreted as a valid Python type,
    a `ValueError` will be raised."""
    if not (value[:1] == '-' or value[:1].isdigit()):  # if not a number
        if value == 'true':
            return True
        if value == 'false':
            return False
        return value
    try:
        return int(value)
    except ValueError:
        return float(value)


def fold_dict(sequence, fold_values=True) -> Dict[str, Any]:
    """Construct a dict out of a sequence of tuples of the form (key, value). If `fold_values` is false, or multiple
    values are given for the same key, those values are collected into a list. If `fold_values` is true (the default),
    the value for keys with only one value (i.e. they appear only once in the sequence) are "folded" into a primitive
    instead of being a list of one element."""
    d = dict() if fold_values else defaultdict(list)

    for key, value in filter(None, sequence):
        if not fold_values or (key in d and type(d[key]) is list):
            d[key].append(value)
        elif key in d:
            d[key] = [d[key], value]
        else:
            d[key] = value
    return d


DELIMITER_KEY_VALUE = '='
DELIMITER_COMMENT = ';'
SECTION_NAME_START = '['
SECTION_NAME_END = ']'



"""
Copyright (C) 2016, 2017, 2020 biqqles.
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
from typing import Dict, List, Tuple, Optional
import os

from dataclassy import Internal

from ..formats import dll
from .. import paths, routines, missions
from . import Entity, EntitySet
from .solars import BaseSolar
from .equipment import Equipment, Commodity
from .ship import Ship
from .goods import EquipmentGood, CommodityGood, ShipPackage


class System(Entity):
    """A star system."""
    file: str
    navmapscale: float = 1.0

    def definition_path(self) -> str:
        """The absolute path to the file that defines this system's contents."""
        return paths.construct_path(os.path.dirname(paths.inis['universe'][0]), self.file)

    def contents(self) -> 'EntitySet[Solar]':
        """All solars in this system."""
        return routines.get_system_contents(self)

    def zones(self) -> 'EntitySet[Zone]':
        """All zones in this system."""
        return self.contents().of_type(Zone)

    def objects(self) -> 'EntitySet[Object]':
        """All objects in this system."""
        return self.contents().of_type(Object)

    def bases(self) -> 'EntitySet[BaseSolar]':
        """All bases in this system."""
        return self.contents().of_type(BaseSolar)

    def planets(self) -> 'EntitySet[Planet]':
        """All planets in this system."""
        return self.contents().of_type(Planet)

    def stars(self) -> 'EntitySet[Star]':
        """All stars in this system."""
        return self.contents().of_type(Star)

    def connections(self) -> 'Dict[Jump, System]':
        """The connections this system has to other systems."""
        return {c: c.destination_system() for c in self.contents().of_type(Jump)}

    def lanes(self) -> 'List[List[TradeLaneRing]]':
        """Return a list of lists of rings, where each nested list represents a complete trade lane and contains each
        ring in that lane in order."""
        rings = EntitySet(c for c in self.contents() if isinstance(c, TradeLaneRing))
        lanes = {r: [] for r in rings if r.prev_ring is None}  # find rings which start a lane
        # group remaining rings into one of these
        for first_ring in lanes:
            current_ring = first_ring
            while current_ring:
                current_ring = rings.get(current_ring.next_ring)
                if current_ring:
                    lanes[first_ring].append(current_ring)
        return [[f, *r] for f, r in lanes.items()]  # flatten grouping dict into list of lists

    def region(self) -> str:
        """The name of the region this system is in, extracted from the infocard."""
        *_, rest = self.infocard('rdl').partition('<TRA data="1" mask="1" def="-2"/><TEXT>')
        region, *_ = rest.partition('</TEXT>')
        return region.title() if region else 'Unknown'


class Base(Entity):
    """A space station or colonised planet, operated by a Faction."""
    ids_info = None  # infocard is defined by the base's solar
    system: str

    def infocard(self, markup='html') -> str:
        """The infocard of this base's solar (Base sections do not define ids_info)."""
        return self.solar().infocard(markup)

    def system_(self) -> System:
        """The entity of the system this base resides in."""
        return routines.get_systems()[self.system]

    def solar(self) -> Optional['BaseSolar']:
        """Confusingly, Freelancer defines bases separately to their physical representation."""
        return self.system_().bases().unique(base=self.nickname)

    def has_solar(self) -> bool:
        """Whether this base has a physical solar."""
        return self.solar() is not None

    def mbase(self) -> Optional[missions.MBase]:
        """The mission base entry for this base."""
        return missions.get_mbases().get(self.nickname)

    def owner(self) -> 'Faction':
        """The faction which owns this base (its IFF)."""
        return self.solar().owner() if self.has_solar() \
            else routines.get_factions()[self.mbase().local_faction] if self.mbase() else None

    def sector(self) -> str:
        """The sector of this base's solar in its system."""
        return self.solar().sector()

    def market(self):
        return routines.get_markets()[self]

    def sells(self) -> Dict['Good', int]:
        """The goods this base sells, of the form {good -> price}."""
        return self.market()[True]

    def buys(self) -> Dict['Good', int]:
        """The goods this base buys, of the form {good -> price}"""
        return self.market()[False]

    def sells_commodities(self) -> Dict[Commodity, int]:
        """The commodities represented by the goods this base sells, mapped to their prices."""
        return {good.commodity(): price for good, price in self.sells().items() if isinstance(good, CommodityGood)}

    def buys_commodities(self) -> Dict[Commodity, int]:
        """The commodities represented by the goods this base buys, mapped to their prices."""
        return {good.commodity(): price for good, price in self.buys().items() if isinstance(good, CommodityGood)}

    def sells_equipment(self) -> Dict[Equipment, int]:
        """The equipment represented by the goods this base sells, mapped to their prices."""
        return {good.equipment_(): price for good, price in self.sells().items() if type(good) is EquipmentGood}

    def sells_ships(self) -> Dict[Ship, int]:
        """The ships represented by the goods this base sells, mapped to their cost."""
        return {good.ship(): good.cost() for good in self.sells() if isinstance(good, ShipPackage)}


class Faction(Entity):
    """A faction, also known as a group, is an organisation in the Freelancer universe, possibly owning bases or
    controlling territory."""
    ids_short_name: int  # resource id for short form name
    rep: Internal[List[Tuple[float, str]]]  # float is between 1 (adored) and -1 (reviled)

    def short_name(self) -> str:
        """The short form of this faction's name."""
        return dll.lookup(self.ids_short_name)

    def bases(self) -> EntitySet[Base]:
        """All bases owned by this faction."""
        return EntitySet(b for s in routines.get_systems() for b in s.bases().where(reputation=self.nickname))

    def rep_sheet(self) -> Dict['Faction', float]:
        """How this faction views other factions - its reputation sheet."""
        factions = routines.get_factions()
        return {factions[faction]: rep for rep, faction in self.rep if faction in factions}

    def can_dock_at(self, base: BaseSolar) -> bool:
        """Whether this faction can dock at the given base."""
        return self.rep_sheet()[base.owner()] > self.NODOCK_REP

    def props(self) -> missions.FactionProps:
        """The FactionProps entry for this faction."""
        return missions.get_faction_props()[self.nickname]

    def legality(self) -> str:
        """The legality of this faction as defined in its FactionProps entry (Lawful or Unlawful)."""
        return self.props().legality.capitalize()

    NODOCK_REP = -0.65


from .solars import Solar, BaseSolar, Jump, Planet, Star, Zone, Object, TradeLaneRing



"""
Copyright (C) 2016, 2017, 2020 biqqles.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

This directory contains functions to read the file formats used by
Freelancer.
"""
import deconstruct as c


class WinStruct(c.Struct):
    __byte_order__ = c.ByteOrder.LITTLE_ENDIAN  # all formats native to Windows are little-endian
